package jq.qde;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import oursland.naming.Identifiable;

import jq.qde.constraint.ConstraintCorrespondence;
import jq.qde.space.QualitativeVariable;
import jq.qde.space.TimeQuantitySpace;
import jq.qde.space.VariableQuantitySpace;
import jq.qde.space.value.MagnitudeInterval;
import jq.qde.space.value.MagnitudeLandmark;
import jq.qde.space.value.QualitativeDirection;
import jq.qde.space.value.QualitativeMagnitude;
import jq.qde.space.value.QualitativeValue;
import jq.qde.space.value.TimeValue;
import jq.qsim.QDEPossibleValues;
import jq.qsim.QSimListener;
import jq.qsim.VariablePossibleValues;

/**
 * @author oursland
 */
public class QDEState implements Identifiable {
	// a set of VariableStates
	private final TimeValue time;
	public final TimeQuantitySpace tspace;
	public QDEQuantitySpace qspace;
	private QualitativeConstraint[] constraints;
	private final HashMap valueMap = new HashMap();
	private final ArrayList predecessors = new ArrayList();
	private final ArrayList successors = new ArrayList();
	private boolean hashCodeSet = false;
	private int hashCode = 0;
	private int id = -1;

	public QDEState(QDEState copy) {
		this(copy.time, copy.qspace, copy.tspace, copy.constraints);
		this.valueMap.putAll(copy.valueMap);
		this.predecessors.addAll(copy.predecessors);
		this.successors.addAll(copy.successors);
	}

	public static QDEState createNextTemplate(TimeValue time, QDEState prevState) {
		return new QDEState(time, prevState);
	}

	private QDEState(TimeValue time, QDEState prevState) {
		this(time, prevState.qspace, prevState.tspace, prevState.constraints);
		addPredecessor(prevState);
	}

	public QDEState(TimeValue time, QualitativeDifferentialEquation qde) {
		this.time = time;
		this.qspace = qde.qspace;
		this.tspace = qde.tspace;
		this.constraints = qde.getConstraints();
		initUnknown();
	}

	public QDEState(
		TimeValue time,
		QDEQuantitySpace qspace,
		TimeQuantitySpace tspace,
		QualitativeConstraint[] constraints) {
		this.time = time;
		this.qspace = qspace;
		this.tspace = tspace;
		this.constraints = (QualitativeConstraint[]) constraints.clone();
		initUnknown();
	}

	private void initUnknown() {
		for( int i = 0; i < this.qspace.getVariableCount(); i++ ) {
			QualitativeVariable var = this.qspace.getVariable(i);
			valueMap.put(var, new QualitativeValue(qspace.getVariableSpace(var)));
		}
	}

	public TimeValue getTime() {
		return time;
	}

	public boolean validateValue(QualitativeValue value) {
		QualitativeVariable var = value.getVariable();
		QualitativeValue prevValue = getValue(var);
		assert prevValue != null;
		if (prevValue.magnitude.isUnknown() && prevValue.direction.isUnknown()) {
			return true;
		} else if (prevValue.magnitude == value.magnitude && prevValue.direction == value.direction) {
			return true;
		} else {
			return false;
		}
	}

	public void print(PrintStream out, String prefix) {
		for (Iterator iter = valueMap.entrySet().iterator(); iter.hasNext();) {
			Map.Entry entry = (Map.Entry) iter.next();
			QualitativeVariable var = (QualitativeVariable) entry.getKey();
			QualitativeValue value = (QualitativeValue) entry.getValue();
			out.print(prefix);
			out.print(var.getName());
			out.print("(");
			time.printSpecification(out, "");
			out.print(")=<");
			value.magnitude.printSpecification(out, "");
			value.direction.printSpecification(out, " ");
			out.print(">");
			out.println();
		}
	}

	public void setVariableMagnitude(QualitativeVariable var, QualitativeMagnitude magnitude) {
		final QualitativeValue value = getValue(var);
		if (magnitude != value.magnitude) {
			valueMap.put(var, new QualitativeValue(qspace.getVariableSpace(var), magnitude, value.direction));
		}
		return;
	}

	public void setVariableDirection(QualitativeVariable var, QualitativeDirection dir) {
		final QualitativeValue value = getValue(var);
		valueMap.put(var, new QualitativeValue(qspace.getVariableSpace(var), value.magnitude, dir));
		return;
	}

	public void setVariableValue(QualitativeVariable var, QualitativeMagnitude magnitude, QualitativeDirection dir) {
		setValue(new QualitativeValue(qspace.getVariableSpace(var), magnitude, dir));
	}

	public void setValue(QualitativeValue value) {
		QualitativeVariable var = value.getVariable();
		if (valueMap.containsKey(var)) {
			valueMap.put(var, value);
		} else {
			throw new Error("Variable \"" + var.getName() + "\" is not in system state.");
		}
	}

	public QualitativeMagnitude getVariableMagnitude(QualitativeVariable var) {
		return getValue(var).magnitude;
	}

	public QualitativeDirection getVariableDirection(QualitativeVariable var) {
		return getValue(var).direction;
	}

	public VariablePossibleValues getInitialValues(QualitativeVariable var) {
		return getValue(var).getInitialValues();
	}

	public VariablePossibleValues getPossibleValues(QualitativeVariable var, QSimListener l) {
		return getValue(var).getPossibleValues(getTime(), l);
	}

	public QDEPossibleValues getInitialValues(QSimListener listener) {
		return QDEPossibleValues.createInitial(this, listener);
	}

	public QDEPossibleValues getPossibleValues(QSimListener listener) {
		return QDEPossibleValues.createNext(this, listener);
	}

	public Iterator stateIterator() {
		return valueMap.values().iterator();
	}

	public int getPredecessorCount() {
		return predecessors.size();
	}

	public QDEState getPredecessor(int i) {
		return (QDEState) predecessors.get(i);
	}

	public void addPredecessor(QDEState state) {
		predecessors.add(state);
	}

	public int getSuccessorCount() {
		return successors.size();
	}

	public QDEState getSuccessor(int i) {
		return (QDEState) successors.get(i);
	}

	public void addSuccessor(QDEState state) {
		successors.add(state);
	}

	public QualitativeValue[] getValues() {
		QualitativeValue[] rc = new QualitativeValue[qspace.getVariableCount()];
		for (int i = 0; i < rc.length; i++) {
			rc[i] = getValue(qspace.getVariable(i));
		}
		return rc;
	}

	public QualitativeValue getValue(QualitativeVariable var) {
		final QualitativeValue rc = (QualitativeValue) valueMap.get(var);
		if (rc != null) {
			assert this.qspace.contains(rc.qspace);
			return rc;
		}
		throw new Error("Variable \"" + var.getName() + "\" not in system state.");
	}

	public int hashCode() {
		if (!hashCodeSet) {
			hashCode = 0;
			for (Iterator iter = valueMap.values().iterator(); iter.hasNext();) {
				QualitativeValue value = (QualitativeValue) iter.next();
				hashCode += hashCode * 4 + value.hashCode();
			}
			hashCodeSet = true;
		}

		return hashCode;
	}

	public boolean isQualitativelyEqual(QDEState otherState) {
		/* The state hashcode generates a number based solely on the  
		 * parameters' magnitude and qdir.  It is used later as a fast  
		 * test during the check for a cycle.  This can be eliminated 
		 * if it isn't worth the (small) overhead.
		 * 
		 * This is now calculated in SystemState.hashCode()
		 */
		if (otherState == this) {
			return true;
		}
		if (otherState == null) {
			return false;
		}
		if (hashCode() != otherState.hashCode()) {
			return false;
		}
		if (this.qspace != otherState.qspace) {
			return false;
		}
		for (Iterator iter = valueMap.keySet().iterator(); iter.hasNext();) {
			QualitativeVariable var = (QualitativeVariable) iter.next();
			if (!getValue(var).isQualitativelyEqual(otherState.getValue(var))) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Returns true if any variable value is diverging.  
	 * @return
	 */
	public boolean isDiverging() {
		for (Iterator iter = valueMap.values().iterator(); iter.hasNext();) {
			QualitativeValue value = (QualitativeValue) iter.next();
			assert this.qspace.contains(value.qspace);
			if (qspace.isDiverging(value)) {
				// should never return true on intervals
				assert !value.magnitude.isInterval();
				return true;
			}
		}
		return false;
	}

	/**
	 * Returns true if all variable value are steady.  
	 * @return
	 */
	public boolean isQuiescent() {
		for (Iterator iter = valueMap.keySet().iterator(); iter.hasNext();) {
			QualitativeVariable var = (QualitativeVariable) iter.next();
			if (!getValue(var).isSteady()) {
				return false;
			}
		}
		return true;
	}

	public boolean createNewLandmarks() {
		boolean changedQspace = false;
		final VariableQuantitySpace[] varspaces = new VariableQuantitySpace[qspace.getVariableCount()];
		for (int i = 0; i < varspaces.length; i++) {
			varspaces[i] = qspace.getVariableSpace(i);
			final QualitativeVariable var = varspaces[i].variable;
			final QualitativeValue value = getValue(var);
			if (value.isNewLandmarkCandidate()) {
				assert qspace.contains(value.qspace);
				// only intervals are candidates for new landmarks
				final MagnitudeInterval interval = (MagnitudeInterval) value.magnitude;
				final MagnitudeLandmark newLM = var.createNewLandmark();
				varspaces[i] = (VariableQuantitySpace) varspaces[i].addAfter(interval.low, newLM);
				setValue(new QualitativeValue(varspaces[i], newLM, value.direction));
				changedQspace = true;
			}
		}
		if (changedQspace) {
			this.qspace = new QDEQuantitySpace(varspaces);
		}
		return changedQspace;
	}

	public boolean constraintValuesArePoints(QualitativeConstraint conp) {
		for (int j = 0; j < conp.getVariableCount(); j++) {
			final QualitativeVariable var = conp.getVariable(j);
			final QualitativeMagnitude magnitude = getValue(var).magnitude;
			/*  Make sure that all arguments are point values.  */
			if (!magnitude.isPoint()) {
				return false;
			}
		}
		return true;
	}

	public ConstraintCorrespondence addCorrespondingValue(QualitativeConstraint constraint) {
		MagnitudeLandmark[] corrValues = new MagnitudeLandmark[constraint.getVariableCount()];
		for (int i = 0; i < corrValues.length; i++) {
			final QualitativeVariable var = constraint.getVariable(i);
			corrValues[i] = (MagnitudeLandmark) getValue(var).magnitude;
		}
		ConstraintCorrespondence corr = new ConstraintCorrespondence(corrValues);
		replaceConstraint(constraint, constraint.addCorrespondingValue(corr));
		return corr;
	}

	private void replaceConstraint(QualitativeConstraint oldCon, QualitativeConstraint newCon) {
		constraints = (QualitativeConstraint[]) constraints.clone();
		for (int i = 0; i < constraints.length; i++) {
			if (constraints[i] == oldCon) {
				constraints[i] = newCon;
			}
		}
	}

	public QualitativeConstraint[] getConstraints() {
		return (QualitativeConstraint[]) constraints.clone();
	}

	public int getVariableCount() {
		return valueMap.size();
	}

	public int getIdentifier() {
		return id;
	}

	public void setIdentifier(int id) {
		this.id = id;
	}

	public int getTimeIndex() {
		return tspace.getQSpaceIndex(getTime());
	}

	public TimeValue getNextTime() {
		int nextTimeQIndex = getTimeIndex() + 1;
		TimeValue time = (TimeValue) tspace.getQValue(nextTimeQIndex);
		return time;
	}
	
	public QualitativeValue getValue(int i) {
		return getValue(qspace.getVariable(i));
	}

	public String toAbbreviation() {
		StringBuffer rc = new StringBuffer();
		rc.append("State " + getIdentifier() + " Time = " + getTime().toAbbreviation());
		return rc.toString();
	}
}
