package jq.qde.constraint;

import jq.qde.QDEState;
import jq.qde.QualitativeConstraint;
import jq.qde.space.QualitativeVariable;
import jq.qde.space.value.QualitativeDirection;
import jq.qde.space.value.QualitativeMagnitude;
import jq.qde.space.value.QualitativeValue;

/**
 * x+y=z
 * @author oursland
 */
public class AddConstraint extends QualitativeConstraint {
	public AddConstraint(QualitativeVariable x, QualitativeVariable y, QualitativeVariable z) {
		super(
			"add",
			new QualitativeVariable[] { x, y, z },
			new ConstraintCorrespondence[] { new ConstraintCorrespondence(x.ZERO, y.ZERO, z.ZERO)});
	}

	private AddConstraint(QualitativeVariable[] args, ConstraintCorrespondence[] corr) {
		super( "add", args, corr );
	}

	public boolean propagateConstraint(QDEState state, MagnitudeDelta change) {
		final QualitativeVariable x = getVariable(0);
		final QualitativeVariable y = getVariable(1);
		final QualitativeVariable z = getVariable(2);

//		final QualitativeVariable var = change.getVariable();

		final QualitativeMagnitude xVal = state.getVariableMagnitude(x);
		final QualitativeMagnitude yVal = state.getVariableMagnitude(y);
		final QualitativeMagnitude zVal = state.getVariableMagnitude(z);
		final QualitativeDirection xDir = state.getVariableDirection(x);
		final QualitativeDirection yDir = state.getVariableDirection(y);
		final QualitativeDirection zDir = state.getVariableDirection(z);
		if (change.isValueChange()) {
			if (xVal.isZero() && yVal.isUnknown() && !zVal.isUnknown()) {
				System.out.println(
					"Addition value propagation from "
						+ z.getName()
						+ " and "
						+ x.getName()
						+ " to "
						+ y.getName()
						+ ".");
				state.setVariableMagnitude(y, zVal);
			} else if (xVal.isUnknown() && yVal.isZero() && !zVal.isUnknown()) {
				System.out.println(
					"Addition value propagation from "
						+ z.getName()
						+ " and "
						+ y.getName()
						+ " to "
						+ x.getName()
						+ ".");
				state.setVariableMagnitude(x, zVal);
			}
		} else if (change.isDirectionChange()) {
			if (isSteadyPropagation(xDir, yDir, zDir)) {
				System.out.println(
					"Addition directional propagation from "
						+ x.getName()
						+ " and "
						+ z.getName()
						+ " to "
						+ y.getName()
						+ ".");
				state.setVariableDirection(y, xDir.reverse());
				return true;
			} else if (isSteadyPropagation(yDir, xDir, zDir)) {
				System.out.println(
					"Addition directional propagation from "
						+ y.getName()
						+ " and "
						+ z.getName()
						+ " to "
						+ x.getName()
						+ ".");
				state.setVariableDirection(x, yDir.reverse());
				return true;
			} else if (isSteadyPropagation(yDir, zDir, xDir)) {
				System.out.println(
					"Addition directional propagation from "
						+ y.getName()
						+ " and "
						+ x.getName()
						+ " to "
						+ z.getName()
						+ ".");
				state.setVariableDirection(z, yDir);
				return true;
			} else if (isSteadyPropagation(xDir, zDir, yDir)) {
				System.out.println(
					"Addition directional propagation from "
						+ x.getName()
						+ " and "
						+ y.getName()
						+ " to "
						+ z.getName()
						+ ".");
				state.setVariableDirection(z, xDir);
				return true;
			} else if (isSteadyPropagation(zDir, xDir, yDir)) {
				System.out.println(
					"Addition directional propagation from "
						+ z.getName()
						+ " and "
						+ y.getName()
						+ " to "
						+ x.getName()
						+ ".");
				state.setVariableDirection(x, zDir);
				return true;
			} else if (isSteadyPropagation(zDir, yDir, xDir)) {
				System.out.println(
					"Addition directional propagation from "
						+ z.getName()
						+ " and "
						+ x.getName()
						+ " to "
						+ y.getName()
						+ ".");
				state.setVariableDirection(y, zDir);
				return true;
			}
		}

		return true;
	}
	
	private static boolean isSteadyPropagation(QualitativeDirection a, QualitativeDirection b, QualitativeDirection c) {
		return a != QualitativeDirection.UNKNOWN && b == QualitativeDirection.UNKNOWN && c == QualitativeDirection.STD;
	}
	
	public boolean initConstraint(QDEState state) {
		return true;
	}
	
	private static final int INC = QualitativeDirection.INC_MASK; 
	private static final int STD = QualitativeDirection.STD_MASK; 
	private static final int DEC = QualitativeDirection.DEC_MASK; 
	private static final int ANY = QualitativeDirection.ANY_MASK; 
	
	private static final int[][] addqdir = {
		{INC, INC, ANY},
		{INC, STD, DEC},
		{ANY, DEC, DEC},
	};
	public static int getQDirMask(QualitativeDirection x, QualitativeDirection y) {
		return addqdir[x.key][y.key];
	}

	public ConstraintValidationResult validateValues(QualitativeValue[] stateArray) {
		/*  Qualitative directions must agree.  */
		final QualitativeDirection d0 = stateArray[0].direction;
		final QualitativeDirection d1 = stateArray[1].direction;
		final QualitativeDirection d2 = stateArray[2].direction;
		int dir = getQDirMask(d0, d1);
		if ((d2.mask != dir) && (dir != ANY)) {
			return new ConstraintValidationResult(false, "qdir mismatch", this, stateArray);
		}

		/*  Loop over all corresponding values for this constraint */
		for( int i = 0; i < getCorrespondingValueCount(); i++ ) {
			final ConstraintCorrespondence corr = getCorrespondingValue(i);
			final QualitativeDirection c0 = stateArray[0].qspace.compare(stateArray[0].magnitude, corr.getValue(stateArray[0]));
			final QualitativeDirection c1 = stateArray[1].qspace.compare(stateArray[1].magnitude, corr.getValue(stateArray[1]));
			final QualitativeDirection c2 = stateArray[2].qspace.compare(stateArray[2].magnitude, corr.getValue(stateArray[2]));
			int rel = getQDirMask(c0, c1);
			if((rel != c2.mask) && (rel != ANY)) {
				return new ConstraintValidationResult(false, "cval mismatch", this, stateArray);
			}
		}
//		VariableQuantitySpace r = stateArray[0].qspace;

		/*  Check infinite values  */
		final QualitativeDirection i0 = stateArray[0].magnitude.getInfiniteSign();
//		VariableQuantitySpace r1 = stateArray[1].qspace;
		final QualitativeDirection i1 = stateArray[1].magnitude.getInfiniteSign();
//		VariableQuantitySpace r2 = stateArray[2].qspace;
		final QualitativeDirection i2 = stateArray[2].magnitude.getInfiniteSign();
		if((getQDirMask(i0, i1) & i2.mask) == 0) {
			return new ConstraintValidationResult(false, "fails infinite check", this, stateArray);
		}
		
		return new ConstraintValidationResult(true, "OK", this, stateArray);
	}

	public QualitativeConstraint createCopy(QualitativeVariable[] args, ConstraintCorrespondence[] corr) {
		return new AddConstraint(args, corr);
	}
}
