package jq.qde.space;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.ListIterator;
import jq.qde.space.value.MagnitudeInterval;
import jq.qde.space.value.MagnitudeLandmark;
import jq.qde.space.value.QualitativeDirection;
import jq.qde.space.value.QualitativeMagnitude;
import jq.qde.space.value.QualitativeValue;
import jq.qde.space.value.TimeInterval;
import jq.qde.space.value.TimeLandmark;

/**
 * @author oursland
 */
public class TimeQuantitySpace implements QuantitySpace {
	public final TimeVariable var;
	private final ArrayList landmarkArray = new ArrayList();
	private final ArrayList intervalArray = new ArrayList();

	public TimeQuantitySpace(TimeVariable var) {
		this(var, new MagnitudeLandmark[] {});
	}
	
	public TimeQuantitySpace(TimeVariable var, MagnitudeLandmark[] lms) {
		this.var = var;
		this.landmarkArray.addAll(Arrays.asList(lms));
		for (int i = 0; i < lms.length-1; i++) {
			this.intervalArray.add(var.getInterval(lms[i], lms[i+1]));
		}
	}

	public QuantitySpace addAfter(MagnitudeLandmark prevLM, MagnitudeLandmark newLM) {
		int index = landmarkArray.indexOf(prevLM)+1;
		add(index, newLM);
		return this;
	}

	private void add(int index, MagnitudeLandmark newLM) {
		landmarkArray.add(index, newLM);
		
		if( index > 0 ) {
			if( index-1 < intervalArray.size() ) {
				intervalArray.set(index-1, var.getInterval(getLandmark(index-1), newLM));
			} else {
				intervalArray.add(var.getInterval(getLandmark(index-1), newLM));
			}
		}
		if( index < landmarkArray.size()-1 ) {
			intervalArray.add(index, var.getInterval(getLandmark(index+1), newLM));
		}	
	}
	
	
	public Iterator landmarkIterator() {
		return landmarkArray.iterator();
	}

	public ListIterator landmarkListIterator() {
		return landmarkArray.listIterator();
	}

	private MagnitudeLandmark getLandmark(int i) {
		return (MagnitudeLandmark) landmarkArray.get(i);
	}

	private MagnitudeInterval getInterval(int i) {
		return (MagnitudeInterval) intervalArray.get(i);
	}

	public MagnitudeLandmark getLandmarkValue(String name) {
		for (int i = 0; i < landmarkArray.size(); i++) {
			if( getLandmark(i).getName().equals(name) ) {
				return getLandmark(i);
			}
		}
		throw new Error("Landmark " + name + " does not exist in variable " + var.getName());
	}
	
	public boolean lessThan(MagnitudeLandmark value, MagnitudeLandmark high) {
		if( value == high ) {
			return false;
		}
		Iterator iter = landmarkListIterator(); 
		while(iter.hasNext()) {
			MagnitudeLandmark lm = (MagnitudeLandmark) iter.next();
			if( lm == value ) {
				break;
			}
			if( lm == high ) {
				return false;
			}
		}
		while(iter.hasNext()) {
			MagnitudeLandmark lm = (MagnitudeLandmark) iter.next();
			if( lm == high ) {
				return true;
			}
		}		
		return false;
	}
	
	public int getQValueCount() {
		return 2*landmarkArray.size() - 1;
	}
	
	public QualitativeMagnitude getQValue( int index ) {
		int lowIndex = index/2;
		if( index%2 == 0 ) {
			// landmark value
			if( lowIndex >= landmarkArray.size() ) {
				add(lowIndex, var.getLandmark(var.createName(lowIndex)));
			}
			return getLandmark(lowIndex);
		} else {
			// interval value
			if( lowIndex >= intervalArray.size() ) {
				add(lowIndex+1, var.getLandmark(var.createName(lowIndex+1)));
			}
			return getInterval(lowIndex);
		}
	}

	public void printSpecification(PrintStream out, String prefix) {
		out.print(prefix);
		out.print( "(" + var.getName() + " (");
		for (Iterator iter = landmarkIterator(); iter.hasNext();) {
			MagnitudeLandmark lm = (MagnitudeLandmark) iter.next();
			out.print(lm.getName());			
			if( iter.hasNext() ) {
				out.print(" ");
			}
		}
		out.print("))");
	}

	public QualitativeDirection compare(QualitativeMagnitude magnitude, QualitativeMagnitude landmark) {
		if( magnitude == landmark ) {
			return QualitativeDirection.ZERO;
		} else if( lessThan(magnitude, landmark) ) {
			return QualitativeDirection.NEGATIVE;
		} else if( greaterThan(magnitude, landmark) ) {
			return QualitativeDirection.POSITIVE;
		}
		return QualitativeDirection.UNKNOWN;
	}

	private boolean lessThan(QualitativeMagnitude low, QualitativeMagnitude high) {
		if( getQSpaceIndex(low) < getQSpaceIndex(high) ) {
			return true;
		} else {
			return false;
		}
	}

	private boolean greaterThan(QualitativeMagnitude low, QualitativeMagnitude high) {
		if( getQSpaceIndex(low) > getQSpaceIndex(high) ) {
			return true;
		} else {
			return false;
		}
	}

	public QualitativeDirection getSign(QualitativeMagnitude magnitude) {
		return QualitativeDirection.UNKNOWN;
	}

	public boolean isNegative(QualitativeMagnitude mag) {
		return getSign(mag).isDecreasing();
	}

	public boolean isPositive(QualitativeMagnitude mag) {
		return getSign(mag).isIncreasing();
	}

	public boolean isLowEnd(QualitativeMagnitude magnitude) {
		return magnitude == landmarkArray.get(0);
	}

	public boolean isHighEnd(QualitativeMagnitude magnitude) {
		return magnitude == landmarkArray.get(landmarkArray.size()-1);
	}

	private QualitativeMagnitude getNext(QualitativeMagnitude mag) {
		return getQValue(getQSpaceIndex(mag)+1);
	}

	private QualitativeMagnitude getPrev(QualitativeMagnitude mag) {
		return getQValue(getQSpaceIndex(mag)-1);
	}

	public QualitativeMagnitude shiftMagnitude(QualitativeMagnitude mag, int i) {
		while( i > 0 ) {
			mag = getNext(mag);
			i--;
		}
		while( i < 0 ) {
			mag = getPrev(mag);
			i++;
		}
		return mag;
	}

	public int getQSpaceIndex(QualitativeMagnitude mag) {
		if( mag instanceof TimeLandmark ) {
			return 2*((TimeLandmark)mag).getTime();
		} else if( mag instanceof TimeInterval ) {
			return getQSpaceIndex( ((TimeInterval)mag).low)+1;
		}
		throw new Error("Unknown time value.");
	}

	public boolean isDiverging(QualitativeValue v) {
		final boolean divergingDown =  isLowEnd(v.magnitude) && v.direction.isDecreasing();
		final boolean divergingUp =   isHighEnd(v.magnitude) && v.direction.isIncreasing();
		return divergingDown || divergingUp;
	}

	public boolean isEnd(QualitativeMagnitude mag) {
		return isHighEnd(mag) && isLowEnd(mag);
	}
}
