package jq.qsim;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import oursland.naming.UIDGenerator;
import jq.qde.QDEQuantitySpace;
import jq.qde.QDEState;
import jq.qde.QualitativeConstraint;
import jq.qde.constraint.ConstraintCorrespondence;
import jq.qde.constraint.ConstraintValidationResult;
import jq.qde.space.QualitativeVariable;
import jq.qde.space.TimeQuantitySpace;
import jq.qde.space.value.QualitativeValue;
import jq.qsim.globalfilter.ChangeFilter;
import jq.qsim.globalfilter.CycleFilter;
import jq.qsim.globalfilter.GlobalFilter;

/**
 * @author oursland
 */
public class QSimFilter {
	public final QDEPossibleValues possibleValues;
	private final QDEState previousState;
	private final TimeQuantitySpace tspace;
	private final QDEQuantitySpace qspace;
	private final QualitativeConstraint[] constraints;

	private final HashMap tupleMap = new HashMap();
	private final ConstraintStatisticMap valueStatistics = new ConstraintStatisticMap();
	private final ConstraintStatisticMap variableStatistics = new ConstraintStatisticMap();
	private boolean tuplesCreated = false;
	private boolean tuplesFiltered = false;

	public QSimFilter(QDEPossibleValues ivals, QualitativeConstraint[] constraints) {
		this.possibleValues = ivals;
		this.previousState = null;
		this.qspace = ivals.getQuantitySpace();
		this.tspace = ivals.getTimeSpace();
		this.constraints = constraints;
		initVariableStatistics(constraints);
	}

	public QSimFilter(QDEState previousState, QSimListener listener) {
		this.possibleValues = previousState.getPossibleValues(listener);
		this.previousState = previousState;
		this.qspace = previousState.qspace;
		this.tspace = previousState.tspace;
		this.constraints = previousState.getConstraints();
		initVariableStatistics(constraints);
	}

	private void initVariableStatistics(QualitativeConstraint[] constraints) {
		for (int i = 0; i < constraints.length; i++) {
			for (int j = 0; j < constraints[i].getVariableCount(); j++) {
				QualitativeVariable var = constraints[i].getVariable(j);
				variableStatistics.markContextUse(constraints[i], var);
			}
			variableStatistics.completeContextStatistics(constraints[i], possibleValues);
		}
		//		System.out.println("Checking variable statistics...");
		//		variableStatistics.printStatistics(System.out);
		//		System.out.println("Done");
		//		System.out.println();
	}

	public void createTuples(QSimListener l) {
		if (!tuplesCreated) {
			tuplesCreated = true;
			for (int i = 0; i < constraints.length; i++) {
				final List resultList = createTuples(constraints[i], l);
				tupleMap.put(constraints[i], resultList);
				valueStatistics.completeContextStatistics(constraints[i], possibleValues);
				//				valueStatistics.printStatistics(System.out);
			}
		}
	}

	private List createTuples(QualitativeConstraint constraint, QSimListener l) {
		final int variableCount = constraint.getVariableCount();
		final int maxIndex = variableCount - 1;
		final ArrayList resultList = new ArrayList();
		final LinkedList iterStack = new LinkedList();
		final QualitativeValue[] valueArray = new QualitativeValue[variableCount];

		int varIndex = 0;
		Iterator currentIter = stateIterator(constraint.getVariable(varIndex));
		while (true) {
			if (!currentIter.hasNext()) {
				// this variable has no more values. Look at the next value in the previous variable.
				varIndex--;
				if (iterStack.isEmpty()) {
					break;
				}
				currentIter = (Iterator) iterStack.removeLast();
			} else {
				valueArray[varIndex] = (QualitativeValue) currentIter.next();
				valueStatistics.setContext(constraint, valueArray[varIndex]);
				assert this.qspace.contains(valueArray[varIndex].qspace);
				if (!valueStatistics.isValid(valueArray[varIndex])) {
					// this value has been eliminated. Move onto the next value
				} else if (varIndex < maxIndex) {
					// this variable has more values we haven't considered
					// and we have more variables to add to the state value list
					varIndex++;
					iterStack.addLast(currentIter);
					currentIter = stateIterator(constraint.getVariable(varIndex));
				} else {
					ConstraintValidationResult validation = constraint.validateValues(valueArray);
					if (validation.isValid) {
						// this variable has more values we haven't considered
						// and we have a complete set of values for the next possible state
						ValueTuple possibleTuple = new ValueTuple(valueArray);
						setConstraintUse(valueArray, constraint);
						resultList.add(possibleTuple);
						l.createTuple(possibleTuple, constraint);
					} else {
						l.rejectTuple(valueArray, constraint, validation);
					}
				}
			}
		}
		return resultList;
	}

	/**
	 * Side-effect: Removes invalid tuples from the list.
	 * @return
	 */
	public boolean filterTuples(QSimListener l) {
		if (!tuplesFiltered) {
			tuplesFiltered = true;
			createTuples(l);
			for (int i = 0; i < constraints.length; i++) {
				QualitativeConstraint constraint = constraints[i];
				int goodTupleCount = 0;
				Iterator tupleIter = tupleIterator(constraint);
				while (tupleIter.hasNext()) {
					ValueTuple tuple = (ValueTuple) tupleIter.next();
//					if (tuple.constraint == constraint) {
						boolean tupleIsGood = true;
						for (int k = 0; k < constraint.getVariableCount(); k++) {
							final QualitativeValue value = tuple.getValue(k);
							assert this.qspace.contains(value.qspace);
							final int valueUseCount = getValueUseCount(value);
							final int varUseCount = getVariableUseCount(value.variable);
							if (valueUseCount != varUseCount) {
								// the number of constraints using a tuple should be equal to the number of constraints affecting a variable.
								// otherwise, this tuple was rejected by at least one applicable constraint, so the tuple is bad.
								tupleIsGood = false;
								tupleIter.remove();
								break; // this will exit the argument loop
							} else {
								goodTupleCount++;
							}
						}
						if (tupleIsGood) {
							goodTupleCount++;
						}
						l.filterTuple(tuple, constraint, tupleIsGood, this);
//					}
				}
				if (goodTupleCount == 0) {
					l.allFilteredOut(constraint);
					return false;
				}
			}
		}
		return true;
	}

	public int getVariableUseCount(QualitativeVariable constraintVar) {
		return variableStatistics.getUseCount(constraintVar);
	}

	public int getValueUseCount(QualitativeValue tupleValue) {
		return valueStatistics.getUseCount(tupleValue);
	}

	/*
	 *  'formstates' is applied after tuple generation, constraint checking and
	 *  Waltz filtering.  From the remaining tuples it forms all possible
	 *  consistent states.
	 *
	 *  This function is initially called for constraint 0; it selects a tuple
	 *  and then calls itself recursively for constraint 1, which then selects
	 *  a tuple consistent with the first one, and calls itself recursively for
	 *  constraint 2, etc.  Through backtracking, every consistent combination
	 *  of tuples will be generated.  Early pruning in the earliest recursive
	 *  calls reduces the potentially large search space.
	 *
	 *  There is one improvement that could be made to this routine that would
	 *  eliminate a bad worst-case behavior.  The worst-case would occur if the
	 *  lowest-numbered constraints shared no parameters (thus, there would be
	 *  no early pruning).  To eliminate this problem, the recursive calls
	 *  should use a "connected ordering" of the constraints.  This "connected
	 *  ordering" could be generated once by analyzing the constraints, and
	 *  then saved for later use here.  Is this improvement worth the bother?
	 *  Probably not, since constraint models are usually given in "connected
	 *  order" anyway.
	 */
	public void formStates(UIDGenerator gen, List resultList, QSimListener l) {
		// call filterTuples in case it hasn't been called manually. 
		// There is a check to make sure filtering is not done twice.
		filterTuples(l);

		List stateList = formCandidateStates(l);
		validateCandidateStates(gen, l, stateList);
//		notifyFormedStates(stateList, l);
//		installCorrespondingValues(stateList, l);
		resultList.addAll(stateList);
	}
	
	/**
	 *  For each constraint, install its corresponding values into this
	 *  state.  If the constraint's arguments are all point values and are
	 *  not already corresponding values, and the constraint is not d//dt or
	 *  MULT(f,g,h) where h = 0, then add the new corresponding values.
	 */
	private boolean installCorrespondingValues(QDEState state, QSimListener l) {
		boolean newcorr = false;

		/*  Loop over each constraint  */
		nextcon : 
		for (int i = 0; i < constraints.length; i++) {
			final QualitativeConstraint conp = constraints[i];

			/**
			 * Copy all the existing correspondences of this constraint and
			 * adjust any cval indices due to a new landmark in a qspace.
			 * 
			 * Constraints and corresponding values are copied in the state constructor.
			 * Indices are automatically updated by the QuantitySpace.
			 *  
			 * cq prints out each copied corresponding value, so we will too.
			 */
			for (int j = 0; j < conp.getCorrespondingValueCount(); j++) {
				ConstraintCorrespondence corrp = conp.getCorrespondingValue(j);
				l.inheritingCorrespondence(conp, corrp);
			}
			/**  
			 * The correspondences for this constraint have been installed.
			 * Now see if a new correspondence should be created based on
			 * the parameter values in this state.
			 */

			/**  
			 * Constraint specific tests to see if new values should be created.
			 * Derivatives don't have corresponding values, so skip them.  
			 * If MULT(f,g,h) and h==0 then don't create new correspondence.
			 * All of the state values for a constraint must be points.
			 * There should not already be a corresponding value for the state values.
			 */
			if( conp.shouldCreateNewCorrespondence(state, l) ) {
				l.addingCorrespondence(conp);
				/*  create new corresponding value.  */
				ConstraintCorrespondence corr = state.addCorrespondingValue(conp);
//				CorrespondingValue corr = new CorrespondingValue(conp, state);
				/*  Add the new set of corresponding values  */
				newcorr = true;
				l.addedCorrespondence(conp, corr);
			}
		}
		return newcorr;
	}

//	private void notifyFormedStates(List stateList, SimulationListener l) {
//		Iterator iter = stateList.iterator();
//		while (iter.hasNext()) {
//			SystemState candidateState = (SystemState) iter.next();
//			l.formedState(candidateState);
//		}
//	}

	private void validateCandidateStates(UIDGenerator gen, QSimListener l, List stateList) {
		Iterator iter = stateList.iterator();
		while (iter.hasNext()) {
			QDEState candidateState = (QDEState) iter.next();
			GlobalFilter failedFilter = validateGlobalFilters(candidateState, l);
			l.tentativeState(candidateState, failedFilter);
			if (failedFilter != null) {
				iter.remove();
				//				System.out.print("Global filter \"" + failedFilter.getName() + "\" rejects: ");
				//				PrintUtil.printState(System.out, candidateState);
				//				System.out.println();
				//				validateGlobalFilters(candidateState);
			} else {
				candidateState.setIdentifier(gen.next());
				l.formedState(candidateState);
				installCorrespondingValues(candidateState, l);
			}
		}
	}

	private List formCandidateStates(QSimListener l) {
		ArrayList stateList = new ArrayList();

		// add an initial completely unknown state
		if (previousState == null) {
			QDEState state = new QDEState(possibleValues.getTime(), qspace, tspace, constraints);
			stateList.add(state);
		} else {
			QDEState state = QDEState.createNextTemplate(possibleValues.getTime(), previousState);
			assert state.getVariableCount() == previousState.getVariableCount();
			stateList.add(state);
		}

		// Loop over all constraints
		for (int i = 0; i < constraints.length; i++) {
			QDEState[] stateArray = (QDEState[]) stateList.toArray(new QDEState[stateList.size()]);
			stateList.clear();
			// Loop over all tuples for this constraint
			formConstraintCandidates(constraints[i], stateArray, stateList, l);
		}
		
		for (Iterator iter = stateList.iterator(); iter.hasNext();) {
			QDEState candidateState = (QDEState) iter.next();
			candidateState.createNewLandmarks();
		}

		return stateList;
	}

	private void formConstraintCandidates(
		QualitativeConstraint constraint,
		QDEState[] candidateStates,
		ArrayList nextStateList,
		QSimListener l) {
		Iterator tupleIter = tupleIterator(constraint);
		while (tupleIter.hasNext()) {
			// For each tuple for this constraint, create a new state based on the current states
			formTupleCandidates((ValueTuple) tupleIter.next(), candidateStates, nextStateList, l);
		}
	}

	private void formTupleCandidates(
		ValueTuple tuple,
		QDEState[] candidateStates,
		ArrayList nextStateList,
		QSimListener l) {
		for (int j = 0; j < candidateStates.length; j++) {
			final QDEState candidateState = candidateStates[j];
			assert candidateState.getVariableCount() >= tuple.size();
			if (!addGoodState(candidateState, tuple, nextStateList, l)) {
				l.mismatchState(tuple, candidateState);
			}
		}
	}

	private boolean addGoodState(QDEState candidateState, ValueTuple tuple, ArrayList nextStateList, QSimListener l) {
		final boolean isGoodState = validateTuple(tuple, candidateState);
		if (isGoodState) {
			nextStateList.add(createTupleState(tuple, candidateState, l));
		}
		return isGoodState;
	}

	private final static GlobalFilter changeFilter = new ChangeFilter();
//	final static GlobalFilter quiescentFilter = new QuiescentFilter();
//	final static GlobalFilter divergenceFilter = new DivegenceFilter();
	private final static GlobalFilter cycleFilter = new CycleFilter();
	private final static GlobalFilter[] globalFilters = { 
		changeFilter, 
//		quiescentFilter, 
//		divergenceFilter, 
		cycleFilter 
	};

	/**
	 * Checks candidate states created from the filtered tuples. Candidate 
	 * states are checked against all states filters ("no change" and "cycle" filters).
	 * Surviving states that are not quiescent and not divergent are returned. 
	 *
	 * The task of checking and installing a candidate state is potentially
	 * time-consuming, so 'formNewStates' does the following:
	 *
	 * (1)  It does as much as possible "in parallel".  For example, as it is
	 * iterating through the set of parameters tentatively installing
	 * their values in a new state, it is also testing for "no change",
	 * divergence, and quiescence, and installing new landmarks.
	 * 
	 * (2)  It tries to eliminate logically unnecessary tests.  For example,
	 * if no new landmarks were created then check for a cycle, but don't
	 * bother checking for new correspondences.
	 * 
	 * @return An array of states that satisfy all constraints and pass all filters. 
	 */
	private GlobalFilter validateGlobalFilters(QDEState candidateState, QSimListener l) {
		for (int i = 0; i < globalFilters.length; i++) {
			if (!globalFilters[i].passesFilter(candidateState, previousState)) {
				return globalFilters[i];
			}
		}

		/*  Candidate state has survived all tests, so make it official.  */
		if (previousState != null) {
			previousState.addSuccessor(candidateState);
		}

		return null;
	}

	private QDEState createTupleState(ValueTuple tuple, QDEState state, QSimListener l) {
		QDEState rc = new QDEState(state);
		for (int i = 0; i < tuple.size(); i++) {
//			l.assignTupleValue()
			QualitativeValue newValue = tuple.getValue(i);
			QualitativeValue oldValue = rc.getValue(newValue.variable);
			if( newValue != oldValue ) {
				rc.setValue(newValue);
				l.assignTupleValue(tuple, newValue);
			}
		}
		return rc;
	}

	private boolean validateTuple(ValueTuple tuple, QDEState state) {
		for (int i = 0; i < tuple.size(); i++) {
			if (!state.validateValue(tuple.getValue(i))) {
				return false;
			}
		}
		return true;
	}

	public Iterator stateIterator(QualitativeVariable var) {
		return possibleValues.valueIterator(var);
	}

	public Iterator tupleIterator(QualitativeConstraint constraint) {
		return ((List) tupleMap.get(constraint)).iterator();
	}

	//	public Iterator statisticIterator() {
	//		return stateStats.values().iterator();
	//	}

	public void printTupleStatistics(PrintStream out, ValueTuple tuple) {
//		final QualitativeConstraint constraint = tuple.constraint;
		for (int i = 0; i < tuple.size(); i++) {
			QualitativeValue value = tuple.getValue(i);
			out.print("[");
			out.print(variableStatistics.getUseCount(value.variable));
			out.print(" ");
			out.print(valueStatistics.getUseCount(value));
			out.print("]\t");
		}
	}

	public boolean isValid(QualitativeValue state) {
		return valueStatistics.isValid(state);
	}

	public void setConstraintUse(QualitativeValue state, QualitativeConstraint constraint) {
		valueStatistics.markContextUse(constraint, state);
	}

	public void setConstraintUse(QualitativeValue[] stateArray, QualitativeConstraint constraint) {
		for (int i = 0; i < stateArray.length; i++) {
			setConstraintUse(stateArray[i], constraint);
		}
	}

	public Iterator constraintIterator() {
		return Arrays.asList(constraints).iterator();
	}
}
